import numpy as np
from matplotlib import mlab
import logging

from signum.tools.scale_manager import ScaleManager

from signum.core.base_container import SignalContainer, SCA
from signum.core.freq_domain_signal import FreqDomainSignal


logger = logging.getLogger(__name__)


class TimeDomainSignal(SignalContainer):
    ATTRIBUTES = SignalContainer.ATTRIBUTES + [SCA('_t_sampling', 1, True)]

    def __new__(cls, *args, f_sampling: float = 1, f_sampling_unit='Hz', **kwargs):
        kwargs['x_unit'] = kwargs.pop('x_unit', 's')
        kwargs['x_description'] = kwargs.pop('x_description', 'Time')
        obj = super().__new__(cls, *args, **kwargs)

        # set sampling time
        f_sampling, f_sampling_unit = ScaleManager.rescale_to_basic(f_sampling, f_sampling_unit)
        obj.t_sampling = 1./f_sampling  # using the t sampling setter

        return obj

    def __repr__(self):
        return super().__repr__() + f' sampled at {ScaleManager.display(self.f_sampling, "Hz")}'

    @property
    def resolution(self):
        return self._base_resolution * self._t_sampling

    @resolution.setter
    def resolution(self, val):
        self._resolution_setter_(val)

    def _resolution_setter_(self, val):
        self._t_sampling = val
        self._resolution_defaults_setter()

    @property
    def t_sampling(self):
        return self.resolution

    @t_sampling.setter
    def t_sampling(self, val):
        self._resolution_setter_(val)

    @property
    def f_sampling(self):
        return 1./self.t_sampling

    @f_sampling.setter
    def f_sampling(self, val):
        self._resolution_setter_(1./val)

    @staticmethod
    def check_metadata(arr1, arr2, require_type=False):
        for i, arr in enumerate([arr1, arr2]):
            if not isinstance(arr, TimeDomainSignal):
                m = f'Input array #{i+1} is not an instance of TimeDomainSignal'
                if require_type:
                    raise TypeError(m)
                logger.debug(m)
                return

        if not np.isclose(arr1.f_sampling, arr2.f_sampling):
            raise ValueError(f"Inputs sampling frequencies mismatch: {arr1.f_sampling:.2E} and {arr2.f_sampling:.2E}")

    @staticmethod
    def csd(x, y=None, **kwargs) -> FreqDomainSignal:
        """Cross spectral density"""

        y_in = y if y is not None else x

        # check that the operation is applicable to the inputs
        TimeDomainSignal.check_metadata(x, y_in, require_type=True)

        # calculate the spectrum
        spectrum, frequencies = mlab.csd(y_in, x, **kwargs, Fs=x.f_sampling)

        # cast type to FreqDomainSingal
        if y is None:
            description = f"{x.description}: power spectral density"
            meta = {**x.meta, 'original_description': x.description}
        else:
            description = f"{x.description} & {y.description}: cross spectral density"
            meta = {**x.meta, **y.meta, 'original_descriptions': [x.description, y.description]}

        spectrum = FreqDomainSignal(spectrum, f_resolution=frequencies[1] - frequencies[0], x_start=frequencies[0],
                                    description=description, meta=meta, unit=x.unit)

        return spectrum

    @staticmethod
    def psd(x, **kwargs):
        """Power spectral density"""

        return TimeDomainSignal.csd(x, **kwargs)


if __name__ == '__main__':
    data1 = TimeDomainSignal(np.random.rand(12), f_sampling=40e6)
    data2 = TimeDomainSignal(np.arange(12).reshape(1, -1), f_sampling=1, unit='V')
    data3 = TimeDomainSignal(np.arange(12), f_sampling=40e6)

    data1.display()

    psd = TimeDomainSignal.psd(data1)
    psd.display()

    csd = TimeDomainSignal.csd(data1, data3)
    csd.display()



import numpy as np

from signum.core.base_container import SignalContainer, SCA


class FreqDomainSignal(SignalContainer):
    ATTRIBUTES = SignalContainer.ATTRIBUTES + [SCA('_f_resolution', 1, True)]

    # alias for the property
    f_resolution = SignalContainer.resolution

    def __new__(cls, *args, f_resolution: float=None, **kwargs):
        obj = super().__new__(cls, *args, **kwargs, x_unit='Hz', x_description='Frequency')

        if f_resolution is not None:
            obj.f_resolution = f_resolution  # using the property setter

        return obj

    @property
    def resolution(self):
        return self._base_resolution * self._f_resolution

    @resolution.setter
    def resolution(self, val):
        self._resolution_setter_(val)

    def _resolution_setter_(self, val):
        self._f_resolution = val
        self._resolution_defaults_setter()

    @property
    def f_sampling(self):
        return self.resolution

    @f_sampling.setter
    def f_sampling(self, val):
        self._resolution_setter_(val)


if __name__ == '__main__':
    data1 = FreqDomainSignal(np.random.rand(12) + 1j * np.random.rand(12), f_resolution=10, unit='mV',
                             description="Frequency response")
    data2 = FreqDomainSignal(np.arange(12).reshape(1, -1), unit='V', description="H")

    data2.reshape(-1).display()
    data1.display()
    data1.display(complex_plot='bode', db_scale=True)
    data1.display(complex_plot='nyquist')




import numpy as np
import re
import logging
from typing import Union

logger = logging.getLogger(__name__)


class UnitPrefixDict(dict):
    """Dictionary mapping standard prefixes for SI units on corresponding orders of magnitude.

    Additional functionalities implemented:
    - inverse mapping (order -> prefix)
    - matching closest defined prefix (if the order is too small/too big for the defined prefixes)
    """

    def __init__(self, *args, **kwargs):
        super(UnitPrefixDict, self).__init__(*args, **kwargs)

        # verify the type and values
        for key, value in self.items():
            self._check_item(key, value)

        # define the inverse mapping
        self._inverse = {}
        self._update_inverse()

    def _update_inverse(self):
        """Update the inverse mapping (to be called in __init__ and __setitem__)."""

        self._check_unique_orders()
        self._inverse = {v: k for k, v in self.items()}

    def _check_unique_orders(self):
        """Verify that the set of orders (dictionary values) is unique, so that the inverse mapping can be defined."""

        if len(set(self.values())) < len(self):
            raise ValueError(f"Orders must be unique")

    @property
    def inverse(self):
        return self._inverse

    @staticmethod
    def _check_item(key, value):
        """Verify that the prefixes and orders are of proper type and value."""

        if not isinstance(key, str):
            raise TypeError(f"Prefix should be a string (got '{key}' of type '{type(key)}')")

        if not isinstance(value, int):
            raise TypeError(f"Order should be an integer (got '{value}' of type '{type(value)}')")

        UnitPrefixDict._check_order_value(value)

    @staticmethod
    def _check_order_value(order):
        """Verify that the order is divisible by 3."""

        if divmod(order, 3)[1]:
            raise ValueError(f"Order should be an integer multiple of 3 (got {order})")

    def __setitem__(self, key: str, value: int):
        self._check_item(key, value)
        super(UnitPrefixDict, self).__setitem__(key, value)
        self._update_inverse()

    def get_prefix(self, order: int):
        """Get a prefix corresponding to the given order."""

        order = int(order)

        self._check_order_value(order)

        if order not in self.values():
            logger.warning(f"No prefix defined for value order {order}")
            return

        return self._inverse[order]

    def match_prefix(self, order: int):
        """Look for the closest defined prefix corresponding to the given order."""

        prefix = self.get_prefix(order)

        if prefix is None:
            prefix, order = self.match_prefix(order - 3 if order > max(self.values()) else order + 3)

        return prefix, order


class ScaleManager(object):
    # units prefixes
    PREFIX_DICT = UnitPrefixDict({'': 0, 'k': 3, 'M': 6, 'G': 9, 'T': 12,
                                  'm': -3, 'u': -6, 'n': -9, 'p': -12, 'f': -15})

    @staticmethod
    def rescale_data(data, order_change: int):
        """Rescale data with the given order of magnitude (exponent of 10)."""

        if order_change == 0:
            return data
        return data * ScaleManager.scaling_factor_from_order(order_change)

    @staticmethod
    def scaling_factor_from_order(order):
        return 10 ** (-order)

    @staticmethod
    def split_unit(unit: str):
        """Split unit to prefix and base."""

        # split the unit to a prefix and the core part
        unit_split_match = re.search("(?P<prefix>[a-zA-Z]?)(?P<core>[a-zA-Z]+((/[a-zA-Z]+)?))", unit)
        if not unit_split_match:
            raise ValueError(f"Invalid unit format: {unit}")
        prefix = unit_split_match.group('prefix')  # e.g. 'p' in 'pm/s'

        if prefix in ScaleManager.PREFIX_DICT.keys():
            unit_core = unit_split_match.group('core')  # main part of the unit (e.g. 'm/s' in 'pm/s')
        else:
            # prefix is empty or is in fact a part of the unit (e.g. 'H' in 'Hz')
            prefix = ''
            unit_core = unit

        return prefix, unit_core

    @staticmethod
    def rescale_unit(unit: str, order_change: int):
        """Adjust rescaling order and change the unit prefix accordingly.

        Returns
        -------
        unit_rescaled   :   str
            the rescaled unit
        order           :   int
            adjusted order to be applied for rescaling the corresponding data
        """

        if order_change == 0:
            return unit, order_change

        prefix, unit_core = ScaleManager.split_unit(unit)
        unit_order = ScaleManager.PREFIX_DICT[prefix]  # e.g. -12 for 'p'

        # define order for the new prefix
        order_for_prefix, order_remainder = divmod(order_change + unit_order, 3)
        order_for_prefix = 3 * order_for_prefix

        # order for prefix may be changed if there is no matching prefix -> rescale data differently
        prefix_new, order_for_prefix = ScaleManager.PREFIX_DICT.match_prefix(order_for_prefix)

        # assemble the final unit
        unit_rescaled = f"{prefix_new}{unit_core}" + (f" * 1E{order_remainder}" if order_remainder else "")

        return unit_rescaled, order_for_prefix - unit_order

    @staticmethod
    def order_from_unit(unit):
        prefix, _ = ScaleManager.split_unit(unit)
        return ScaleManager.PREFIX_DICT[prefix]  # e.g. -12 for 'p'

    @staticmethod
    def adjust_scale(data):
        """Propose a scaling factor magnitude (exponent of 10, an integer multiple of 3) to best represent the data."""

        max_val = np.nanmax(np.abs(data)).item() or 1  # 0 is not a valid argument for logarithm (next step)
        order = np.log10(max_val)
        order = np.sign(order) * np.floor(np.abs(order))
        return int(3 * (order // 3))

    @staticmethod
    def rescale(data, unit: str, order_change: int = None):
        """Rescale data and its unit. Rescaling order may be provided or automatically adjusted to the data."""

        # define the order
        if order_change is None:
            order_change = ScaleManager.adjust_scale(data)

        # adjust the order to the available prefixes; rescale the unit
        unit_rescaled, order_change = ScaleManager.rescale_unit(unit, order_change)

        # rescale the data
        return ScaleManager.rescale_data(data, order_change), unit_rescaled

    @staticmethod
    def rescale_to_basic(data, unit: str):
        """Rescale data and its unit to base unit, e.g. 3 kHz -> 3000 Hz, 0.2 mm -> 0.0002 m"""

        order_change = - ScaleManager.order_from_unit(unit)
        return ScaleManager.rescale(data, unit, order_change=order_change)

    @staticmethod
    def display(number: Union[float, complex], unit: str, precision: int = 2):
        """Rescale a number and its unit and return a string representation with given decimal precision."""

        n_rescaled, unit_rescaled = ScaleManager.rescale(number, unit)
        return f'{n_rescaled:.{precision}f} {unit_rescaled}'





# based on: https://www.numpy.org/devdocs/user/basics.subclassing.html

import matplotlib.pyplot as plt
import numpy as np
import logging
from collections import abc
from typing import Iterable, List, NamedTuple, Any

from signum.tools.plotting_tools import adjust_plot_scale


logger = logging.getLogger(__name__)

SCA = NamedTuple('DataContainerAttribute', [('name', str), ('default_value', Any), ('is_core', bool)])


class SignalContainer(np.ndarray):
    """A subclass of numpy.ndarray, wrapping array data to include metadata envelope (such as sampling frequency)."""

    ATTRIBUTES = [SCA('unit', '', True),
                  SCA('x_unit', '', True),
                  SCA('_base_resolution', 1, True),
                  SCA('x_start', 0, False),
                  SCA('_nonstandard_x_axis', None, False),
                  SCA('description', '', False),
                  SCA('x_description', '', False),
                  SCA('meta', {}, False)]

    def __new__(cls, data, unit: str = None, x_unit: str = None, description: str = None, x_description: str = None,
                x_start: float = None, x_axis=None, resolution: float = None, meta: dict = None):
        """Create a new instance of a DataContainer.

        Parameters
        ----------
        data            :   array_like
            input data to be converted to an array (see numpy.asarray for more details)
        unit            :   str, optional
            data unit
        x_unit          :   str, optional
            unit of the x axis data (e.g. 's', 'Hz')
        description     :   str, optional
            data description
        x_description   :   str, optional
            description of the x axis data (e.g. 'Time')
        x_start         :   float, optional
            starting value for an automatically-generated x axis (calculated from data length and resolution)
        x_axis          :   array_like, optional
            nonstandard x axis data (e.g. if the signal data is not equally spaced); must match the first dimension
            of the data shape
        resolution      :   float, optional
            data spacing (in the x axis units)
        meta            :   dict, optional
            additional, user-provided metadata
        """

        obj = np.asarray(data).view(cls)

        def gv(val, name):
            return val if val is not None else getattr(obj, name)

        # definition of object properties
        obj.unit = gv(unit, 'unit')
        obj.x_unit = gv(x_unit, 'x_unit')
        obj.description = gv(description, 'description')
        obj.x_description = gv(x_description, 'x_description')
        obj.x_start = gv(x_start, 'x_start')
        obj.resolution = gv(resolution, 'resolution')  # using the resolution setter
        obj.meta = gv(meta, 'meta')

        if x_axis is not None:
            obj.x_axis = x_axis  # using the (nonstandard) x axis setter

        return obj

    def __array_finalize__(self, obj):
        """Finalize DataContainer instance creation by assigning the envelope attributes.

        This method is automatically called in every case a new instance of DataContainer is created (unlike __new__ and
        __init_, used only during explicit constructor call; in that case, __array_finalize__ is called first).
        See https://www.numpy.org/devdocs/user/basics.subclassing.html for more details.

        The goal of using this method is to make sure that all the envelope attributes are defined for the newly created
        object."""

        # update the attributes of the newly created object
        self._update_from_obj(obj)

    def _update_from_obj(self, obj=None):
        """Update self to include all the envelope attributes. Copy values from the template object or use defaults.

        The attribute values can come from three sources, depending on the way an instance is created:
        - class defaults (self.defaults from self.ATTRIBUTES)
        - previously defined values (self.__dict__)
        - template object (obj, instance of the same class as self).
        """

        # create container for attributes to be set
        _attributes = {}

        # pre-fill with default values; make sure all attribute names are present
        _attributes.update(self.defaults)

        # preserve current attribute values
        # (if present; when initializing from explicit constructor call, self.__dict__ at this point is empty)
        _attributes.update(self.__dict__)

        # copy attributes from obj (if it is of the same type as self; it might also be None)
        if isinstance(obj, type(self)):
            _attributes.update({k: getattr(obj, k) for k in _attributes.keys()})

        # update the instance attributes with the final set
        self.__dict__.update(_attributes)

    @property
    def defaults(self):
        """Return a dictionary of the default values for the envelope attributes."""

        return {t.name: t.default_value for t in self.ATTRIBUTES}

    @property
    def core_attributes(self) -> List[str]:
        """Return a list of names of the core envelope attributes.

        These are the attributes that need to be taken into account when performing an operation (e.g. addition) on two
        DataContainers (their values must be the same for both inputs - otherwise the operation is not valid).
        """

        return [t.name for t in self.ATTRIBUTES if t.is_core]

    @property
    def envelope_attributes(self):
        """Return a dictionary of current values of the core envelope attributes."""

        return self.get_attributes(self.core_attributes)

    @property
    def envelope_attributes_extended(self):
        """Return a dictionary of current values of the core envelope attributes."""

        return self.get_attributes(self.defaults.keys())

    def get_attributes(self, attr_names: Iterable[str]):
        return {attr: getattr(self, attr, None) for attr in attr_names}

    @envelope_attributes.setter
    def envelope_attributes(self, envelope_dict):
        self.set_attributes(envelope_dict)

    def set_attributes(self, attributes_dict):
        invalid_keys = set(attributes_dict.keys()) - set(self.defaults.keys())
        n = len(invalid_keys)
        if n:
            raise AttributeError(f"{type(self)} has no attribute{'s' if n > 1 else ''}: {invalid_keys}")

        for k, v in attributes_dict.items():
            setattr(self, k, v)

    def copy_attributes(self, other):
        if not isinstance(other, type(self)):
            logger.warning(f"The template object is not an instance of {self.__class__.__name__}")
            return

        self._update_from_obj(other)

    @staticmethod
    def recast_type(func):
        """Perform view casting and copy attributes from the first input to first output of the decorated function."""

        def wrapper(data_in: SignalContainer, *args, **kwargs):
            data_out = func(data_in, *args, **kwargs)

            # if there are multiple outputs - handle only the first one, leave the rest unchanged
            if isinstance(data_out, tuple):
                data_out, *other_outputs = data_out
            else:
                other_outputs = ()

            # view casting
            data_out = data_out.view(type(data_in))

            # propagating attributes
            if isinstance(data_out, SignalContainer):
                data_out.copy_attributes(data_in)

            # fold the outputs tuple back
            return (data_out, *other_outputs) if other_outputs else data_out

        return wrapper

    def __repr__(self):
        return f"{self.description or ''}{': ' if self.description else ''}{self.__class__.__name__} object " \
               f"containing {self.dtype} data array of shape {self.shape}"

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):

        args_envelope = []

        def check_envelope_attributes(obj, side="Inputs"):
            new_envelope = obj.envelope_attributes
            if not len(args_envelope):
                args_envelope.append(new_envelope)
            else:
                if args_envelope[0] != new_envelope:
                    raise TypeError(f"{side} attributes mismatch! ({args_envelope[0]}, {new_envelope})")

        inputs = []
        for arg in args:
            if isinstance(arg, self.__class__):
                check_envelope_attributes(arg)
                inputs.append(arg.view(np.ndarray))  # cast to plain ndarray
            else:
                inputs.append(arg)

        outputs = kwargs.pop('out', None)
        if outputs:
            out_args = []
            for output in outputs:
                if isinstance(output, self.__class__):
                    check_envelope_attributes(output, 'Input/output')
                    out_args.append(output.view(np.ndarray))
                else:
                    out_args.append(output)
            kwargs['out'] = tuple(out_args)
        else:
            outputs = (None,) * ufunc.nout

        results = super(SignalContainer, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)

        if results is NotImplemented:
            return NotImplemented

        if ufunc.nout == 1:
            results = (results,)

        results_converted = []
        for result, output in zip(results, outputs):
            out = result if output is None else output

            # convert back to DataContainer
            rconv = np.asarray(out).view(self.__class__)
            rconv.copy_attributes(self)
            results_converted.append(rconv)

        return results_converted[0] if len(results_converted) == 1 else tuple(results_converted)

    def __getitem__(self, item):
        res = super().__getitem__(item)

        if isinstance(res, type(self)):
            if isinstance(item, tuple):
                item = item[0]

            if not isinstance(item, slice) or self._nonstandard_x_axis is not None:
                res._nonstandard_x_axis = self.x_axis[item]  # nonstandard choice of indices
            else:
                res.x_start = self.x_axis[item][0] if res.size else (self.x_axis[0] if self.size else 0)
                if item.step is not None:
                    res._base_resolution *= item.step

        return res

    @property
    def resolution(self):
        return self._base_resolution

    @resolution.setter
    def resolution(self, val):
        logger.debug(f"Setting base resolution to {val}")
        self._base_resolution = val

    def _resolution_defaults_setter(self):
        self._base_resolution = 1

    @property
    def x_axis(self):
        if self._nonstandard_x_axis is not None:
            return self._nonstandard_x_axis

        x0 = self.x_start if self.x_start is not None else 0
        return x0 + np.arange(self.shape[-1]) * self.resolution

    @x_axis.setter
    def x_axis(self, val):
        if val is None:
            return

        if not isinstance(val, (abc.Sequence, abc.Iterable)):
            raise TypeError("x axis should be a sequence")

        if not len(val) == self.shape[-1]:
            raise ValueError(f"Length of the x axis ({len(val)}) does not match the signal shape ({self.shape})")

        self._nonstandard_x_axis = np.array(val)

        if val is not None:
            self.x_start = 0

    @property
    def x_label(self):
        return self.x_description + (f' [{self.x_unit}]' if self.x_unit else '')

    @property
    def y_label(self):
        return self.description + (f' [{self.unit}]' if self.unit else '')

    @property
    def magnitude(self):
        mag = np.abs(self)
        mag.description = (self.description or '') + f' (magnitude)'
        return mag

    @property
    def magnitude_db(self):
        mag_db = 20 * np.log10(np.abs(self))  # TODO: dB for power quantities
        mag_db.unit = 'dB'
        mag_db.description = (self.description or '') + f'(magnitude, dB)'
        return mag_db

    @property
    def phase(self):
        return self.get_phase()

    def get_phase(self, rad=False, unwrapped=True):
        """Calculate phase of the signal.

        Parameters
        ----------
        rad         :   bool
            if True, return the phase in radians; otherwise (default) - use degrees
        unwrapped   :   bool
            if True (default), unwrap the phase to avoid 2pi jumps; otherwise, return the phase limited to (-pi, pi)
            or (-180, 180), depending on the desired unit
        """

        ph = np.angle(self)  # phase in radians

        if unwrapped:
            ph = np.unwrap(ph)
        if not rad:
            ph = np.rad2deg(ph)

        ph = ph.view(self.__class__)
        ph.copy_attributes(self)
        ph.unit = 'rad' if rad else 'deg'
        ph.description = (ph.description or '') + ' (phase)'
        return ph

    def display(self, complex_plot: str = 'bode', show: bool = True, title: str = None, **kwargs):
        """Create a plot to display the data.

        Parameters
        ----------
        complex_plot    :   str
            if the data is complex, this argument specifies the type of plot to be created. Available options:
            'bode', 'nyquist', 'iq'. For real-valued data, this argument is ignored.
        show            :   bool
            if True, display the plot; setting to False allows later customisation before displaying it
        title           :   str
            title for the plot; if not provided, the description of the signal is used
        kwargs
            additional keyword arguments passed to the specific plotting methods (or matplotlib.pyplot.plot)
        """

        if self.size < 2:
            kwargs['marker'] = kwargs.get('marker', 'o')  # put a single dot if there is only one data point

        if np.iscomplexobj(self):
            if complex_plot == 'bode':
                fig, axes = self._display_complex_bode(**kwargs)
            elif complex_plot == 'nyquist':
                fig, axes = self._display_complex_nyquist(**kwargs)
            elif complex_plot == 'iq':
                fig, axes = self._display_complex_iq(**kwargs)
            else:
                raise ValueError(f"Unknown complex plot type code: {complex_plot}; "
                                 f"should be one of: 'bode', 'nyquist', 'iq'")
        else:
            fig, axes = self._display_real(**kwargs)

        for ax in axes:
            ax.grid(color='lightgray')

        if complex_plot != 'nyquist':
            adjust_plot_scale(axes[-1], x_unit=self.x_unit, x_label=self.x_description)

        fig.suptitle(title or self.description, fontsize=14)

        if show:
            plt.show()

        return fig, axes

    def _display_complex_bode(self, db_scale: bool = False, rad=False, unwrapped=False, figsize=(8, 6), **kwargs):
        """Display a complex signal in a form of Bode plot (magnitude and phase).

        Parameters
        ----------
        db_scale    :   bool
            if True, display the magnitude in decibels.
        rad         :   bool
            if True, display the phase in radians; otherwise, use degrees
        unwrapped   :   bool
            if True, unwrap the phase to avoid 2pi (or 180 deg) jumps; otherwise, show phase in (-pi, pi) range
            (or (-180, 180) for degrees)
        figsize     :   2-tuple
            figure size (passed to plt.subplots)
        """

        fig, axes = plt.subplots(2, 1, sharex='all', figsize=figsize)

        mag = self.magnitude_db if db_scale else self.magnitude
        axes[0].plot(mag.x_axis, mag.T, **kwargs)
        adjust_plot_scale(axes[0], y_unit=mag.unit if not db_scale else None, y_label="Magnitude")

        phase = self.get_phase(rad=rad, unwrapped=unwrapped)
        axes[1].plot(phase.x_axis, phase.T, **kwargs)
        axes[1].set_ylabel(f"Phase [{phase.unit}]")

        return fig, axes

    def _display_complex_iq(self, sharey='all', figsize=(8, 6), **kwargs):
        """Display a complex signal in two subplots: real and imaginary part separately.

        Parameters
        ----------
        sharey  :   str or bool
            passed to matplotlib.pyplot.subplots; if 'all' (or True), the y axis of both subplots is shared.
        figsize     :   2-tuple
            figure size (passed to plt.subplots)
        """

        fig, axes = plt.subplots(2, 1, sharex='all', sharey=sharey, figsize=figsize)

        axes[0].plot(self.x_axis, self.real.T, **kwargs)
        adjust_plot_scale(axes[0], y_unit=self.unit, y_label="Real part")

        axes[1].plot(self.x_axis, self.imag.T, **kwargs)
        adjust_plot_scale(axes[1], y_unit=self.unit, y_label="Imaginary part")

        return fig, axes

    def _display_complex_nyquist(self, figsize=(6, 6), **kwargs):
        """Display a complex-valued signal in a form of a Nyquist plot (imaginary vs real part)."""

        fig, ax = plt.subplots(figsize=figsize)
        axes = np.array([ax])

        ax.set_aspect('equal')
        ax.plot(self.real.T, self.imag.T, **kwargs)

        ax.axvline(0, color='k', zorder=1)
        ax.axhline(0, color='k', zorder=1)

        adjust_plot_scale(ax, x_unit=self.unit, y_unit=self.unit, x_label="I component", y_label="Q component")

        lim = np.max(np.abs([*ax.get_xlim(), *ax.get_ylim()]))
        ax.set_xlim([-lim, lim])
        ax.set_ylim([-lim, lim])

        return fig, axes

    def _display_real(self, figsize=(8, 5), **kwargs):
        """Display a real-valued signal."""

        fig, ax = plt.subplots(figsize=figsize)
        axes = np.array([ax])

        ax.plot(self.x_axis, self.T, **kwargs)
        adjust_plot_scale(ax, y_unit=self.unit, y_label=self.description)

        return fig, axes



from matplotlib.ticker import FuncFormatter

from signum.tools.scale_manager import ScaleManager


def adjust_axis_scale(ax, which: str, unit=None, label=None):
    if which not in ['x', 'y']:
        raise ValueError(f"'which' must be 'x' or 'y' (got '{which}')")

    if not unit and not label:
        return

    unit_rescaled = ''

    if unit:
        order = ScaleManager.adjust_scale(getattr(ax, f'get_{which}lim')())
        unit_rescaled, order = ScaleManager.rescale_unit(unit, order)
        unit_rescaled = f' [{unit_rescaled}]'
        scale = ScaleManager.scaling_factor_from_order(order)

        tick_formatter = FuncFormatter(lambda x, pos: f"{scale * x:g}")
        getattr(ax, f'{which}axis').set_major_formatter(tick_formatter)

    getattr(ax, f'set_{which}label')(f"{label or ''}{unit_rescaled}")


def adjust_plot_scale(ax, x_unit=None, x_label=None, y_unit=None, y_label=None):
    adjust_axis_scale(ax, 'x', x_unit, x_label)
    adjust_axis_scale(ax, 'y', y_unit, y_label)




# based on: https://packaging.python.org/tutorials/packaging-projects/  (as of 29.09.2019).

import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="signum",
    version="0.1.7",
    author="savetheginger",
    author_email="mysh995@gmail.com",
    description="NumPy-based signal data container tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/savetheginger/signum",
    packages=setuptools.find_packages(exclude=('examples',)),
    python_requires='>=3.7',
    install_requires=[
        'numpy>=1.16',
        'matplotlib'
    ],
    extras_require={
        'examples': ['notebook']  # jupyter notebook
    }
)
